
#include <stdio.h>
#include <ctype.h>
#include <dos.h>
#include <malloc.h>
#include <errno.h>
#include <fcntl.h>
#include "doselks.h"

#define MK_FP(Seg,Off) ((char far*)( ((long)Seg << 16) +Off))
/* movedata(sseg, soff, dseg, doff, len) */

void build_stack(/* char ** argv, char ** envp */);

long elks_int_orig;
union  REGS regs;
unsigned int memory;
int process_running;
int exit_code = 255;

main(argc, argv, envp)
int argc;
char ** argv;
char ** envp;
{
   int mode;

   current_ds = __get_ds();
   memory = __segalloc(8192);	/* Allocate 8192 paragraphs */
   if( memory == 0 )
   {
      printf("Unable to allocate 128k for executable\n");
      exit(exit_code);
   }

   if( argc <= 1 )
   {
      printf("Usage: %s <8086_exe> [args]\n", argv[0]);
      exit(exit_code);
   }

   if( loadexe(argv[1]) >= 0 )
   {
      build_stack(argv+1, envp);

      elks_int_orig = __getvect(ELKSINT);
      __setvect(ELKSINT, (long)elks_int);

#ifndef MAKEME
      run_executable();
#else
      dump_core();
#endif

      __setvect(ELKSINT, elks_int_orig);
   }
   else
   {
      printf("Cannot execute %s\n", argv[1]);
   }
   return(exit_code);
}

static int loadexe(fname)
char * fname;
{
	char bbuf[1024];
	int i, fd;
	unsigned int ssize, size;
	unsigned destseg;
	struct elks_exec_hdr mh;
	unsigned char *dsp;

	if( ( fd = open(fname, O_RDONLY|O_BINARY) ) < 0 )
	{
		return -ENOEXEC;
	}

	memset(bbuf, '\0', 1024);

	/* Zap the memory first */
	destseg = memory;
	for(i=0; i<128; i++)
	   movedata(current_ds, (unsigned)bbuf, destseg+i*64, 0, 1024);

	/* Load the elks binary image and set it up in a suitable VM86
	/* segment. Load CS and DS/SS according to image type. chmem
	/* is ignored we always use 64K segments */

	if(read(fd, &mh,sizeof(mh))!=sizeof(mh))
	{
		return -ENOEXEC;
	}
	if(mh.hlen!=EXEC_HEADER_SIZE)
	{
		return -ENOEXEC;
	}
	if(mh.type!=ELKS_COMBID&&mh.type!=ELKS_SPLITID)
	{
		return -ENOEXEC;
	}

#ifdef DEBUGGER
	fprintf(stderr,"Executable - %lX. tseg=%ld dseg=%ld bss=%ld\n",
		mh.type,mh.tseg,mh.dseg,mh.bseg);
#endif

	elks_regs.cs = destseg;

	if(mh.type==ELKS_COMBID)
		ssize = mh.tseg+mh.dseg;
	else
		ssize = mh.tseg;

	for(size=0; size<ssize; )
	{
		i = sizeof(bbuf); if( i>ssize-size ) i=ssize-size;

		if(read(fd,bbuf,i)!=i)
		{
			return -ENOEXEC;
		}
		
	   	movedata(current_ds, (unsigned)bbuf, destseg, size, i);
	   	size += i;
	}

	if(mh.type==ELKS_SPLITID)
	{
		ssize = mh.dseg;

		destseg += 4096; /* 65536/16 */
		for(size=0; size<ssize; )
		{
			i = sizeof(bbuf); if( i>ssize-size ) i=ssize-size;

			if(read(fd,bbuf,i)!=i)
			{
				return -ENOEXEC;
			}
		
	   		movedata(current_ds, (unsigned)bbuf, destseg, size, i);
	   		size += i;
		}
	}

	/* memset(dsp+mh.dseg,0, mh.bseg); */

	close(fd);
	/*
	 *	Load the VM86 registers
	 */
	 
	/* elks_regs.cs= Done above; */

	elks_regs.ds=destseg;
	elks_regs.ss=destseg;   /* Same as DS */

	elks_regs.es=0;		/* This will be preserved */
	elks_regs.sp=65534; 	/* Args stacked later */
	elks_regs.pc=0;		/* Run from 0 */
	
	/*
	 *	Loaded
	 */

	return 0;
}

void
build_stack(argv, envp)
char ** argv; char ** envp;
{
	char buffer[1024];

	char **p;
	int argv_len=0, argv_count=0;
	int envp_len=0, envp_count=0;
	int stack_bytes;
	unsigned short * pip;
	char * pcp;

	/* How much space for argv */
	for(p=argv; *p; p++)
	{
	   argv_count++; argv_len += strlen(*p)+1;
	}

	/* How much space for envp */
	for(p=envp; *p; p++)
	{
	   envp_count++; envp_len += strlen(*p)+1;
	}

	/* tot it all up */
	stack_bytes = 2				/* argc */
	            + argv_count * 2 + 2	/* argv */
		    + argv_len
		    + envp_count * 2 + 2	/* envp */
		    + envp_len;

	if( stack_bytes > sizeof(buffer)-2 )
	{
		printf("Environment too large\n");
		exit(exit_code);
	}
	/* Allocate it */
	elks_regs.sp -= stack_bytes;

/* Sanity check 
	printf("Argv = (%d,%d), Envp=(%d,%d), stack=%d\n",
	        argv_count, argv_len, envp_count, envp_len, stack_bytes);

*/
	/* Now build the data */
	pip=(unsigned short *) buffer;
	pcp=buffer+sizeof(unsigned short *)*(1+argv_count+1+envp_count+1);

	*pip++ = argv_count;
	for(p=argv; *p; p++)
	{
	   *pip++ = (unsigned short) pcp;
	   strcpy(pcp, *p);
	   pcp += strlen(*p)+1;
	}
	*pip++ = 0;

	for(p=envp; *p; p++)
	{
	   *pip++ = (unsigned short) pcp;
	   strcpy(pcp, *p);
	   pcp += strlen(*p)+1;
	}
	*pip++ = 0;
	movedata(current_ds, (unsigned)buffer,
                 elks_regs.ds, elks_regs.sp, stack_bytes);
}

/***************************************************************************/

dump_core()
{
	char bbuf[1024];
	int i, fd;
	unsigned destseg;
	printf("Dumping core\n");

	if( ( fd = open("core", O_WRONLY|O_BINARY|O_CREAT|O_TRUNC) ) < 0 )
	{
		return -1;
	}

	destseg = memory;

	for(i=0; i<128; i++)
	{
	   movedata( destseg+i*64, 0, current_ds, (unsigned)bbuf, 1024);
	   if( write(fd, bbuf, 1024) < 0 )
	      printf("FAIL 7\n");
	}

	close(fd);
}

